(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2014, 2015)
Author:  Harald Zankl (2014, 2015)
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2014, 2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2014, 2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Rule_Labeling_Impl
imports
  Decreasing_Diagrams2
  Critical_Pairs_Impl
  Equational_Reasoning_Impl
  Auxx.Map_Choice
begin

abbreviation DD2_last where "DD2_last \<equiv> Decreasing_Diagrams2.last"

context
  fixes ren :: "'v :: infinite renaming2" 
begin

definition critical_peaks_impl :: "('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> (bool \<times> ('f,'v) lpeak) list" 
  where "critical_peaks_impl P R \<equiv> concat (map (\<lambda> (l,r). 
    concat (map (\<lambda> p. let C = ctxt_of_pos_term p l; l'' = l |_ p; b = (C = \<box>) in 
      if is_Var l'' then [] else 
      concat (map (\<lambda> (l',r'). case mgu_vd ren l'' l' of 
        Some (\<sigma>,\<tau>) \<Rightarrow> [(b, (l \<cdot> \<sigma>, (l, r), [], \<sigma>, True, r \<cdot> \<sigma>), (l \<cdot> \<sigma>, (l', r'), p, \<tau>, True, (C \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>))] 
      | None \<Rightarrow> []) R)) (poss_list l))) P)"

lemma critical_peaks_impl: 
  "set (critical_peaks_impl R R) = critical_peaks ren (set R)" (is "?l = ?r")
proof -
  note cpdefs = critical_peaks_impl_def critical_peaks_def set_concat
    set_map Let_def poss_list_sound
  {
    fix b s1 s2
    assume "(b,s1,s2) \<in> ?r"
    from this[unfolded cpdefs]
    obtain l r l' r' l'' C \<sigma> \<tau> where
      b: "b = (C = \<box>)" and s1: "s1 = (l \<cdot> \<sigma>, (l, r), [], \<sigma>, True, r \<cdot> \<sigma>)" 
      and s2: "s2 = ( l \<cdot> \<sigma>, (l', r'), hole_pos C, \<tau>, True, (C \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>)"
      and lr: "(l, r) \<in> set R" and lr': "(l', r') \<in> set R"
      and l: "l = C\<langle>l''\<rangle>" and l'': "is_Fun l''"
      and mgu: "mgu_vd ren l'' l' = Some (\<sigma>, \<tau>)"
      by auto
    let ?p = "hole_pos C"
    from l have p: "?p \<in> poss l" by auto
    from l p have C: "C = ctxt_of_pos_term ?p l" by auto
    from l p have l: "l |_ ?p = l''" by auto
    have "(b,s1,s2) \<in> ?l" unfolding cpdefs b s1 s2
      by (rule, rule, rule lr, unfold cpdefs, rule, rule imageI[OF p],
      insert C l lr' l'' mgu, force)
  } 
  then have "?r \<subseteq> ?l" by auto
  moreover
  {
    fix b s1 s2
    assume "(b,s1,s2) \<in> ?l"
    from this[unfolded cpdefs]
    obtain l r p l' r' \<sigma> \<tau> where
      lr: "(l,r) \<in> set R" and p: "p \<in> poss l" and lp: "is_Fun (l |_ p)" 
      and lr': "(l',r') \<in> set R" 
      and mgu: "Some (\<sigma>,\<tau>) = mgu_vd ren (l |_ p) l'" 
      and b: "b = (ctxt_of_pos_term p l = \<box>)" 
      and s1: "s1 = (l \<cdot> \<sigma>, (l, r), [], \<sigma>, True, r \<cdot> \<sigma>)" 
      and s2: "s2 = (l \<cdot> \<sigma>, (l', r'), p, \<tau>, True, (ctxt_of_pos_term p l \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>)"  
    by force  
    from p have pp: "hole_pos (ctxt_of_pos_term p l) = p" by simp
    from s1 have s: "get_target s1 = r \<cdot> \<sigma>" unfolding get_target_def by simp    
    from s2 have t: "get_target s2 = (ctxt_of_pos_term p l \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>" 
      unfolding get_target_def by simp    
    from critical_peaksI[OF lr lr' _ lp mgu[symmetric] s t b] 
    have "(b,s1,s2) \<in> ?r" using  ctxt_supt_id[OF p, symmetric] s1 s2 pp
    unfolding get_target_def by auto
  }
  then have "?l \<subseteq> ?r" by auto
  ultimately show ?thesis by auto
qed
end

type_synonym ('f, 'v) rule_lab_repr = "(('f, 'v) rule \<times> nat) list"

definition rule_lab_repr_to_lab :: "('f :: compare_order, 'v :: compare_order ) rule_lab_repr \<Rightarrow> ('f, 'v) rule \<Rightarrow> nat"
where
  "rule_lab_repr_to_lab ps = fun_of_map (ceta_map_of ps) 0"

fun rseq_to_step_list :: "('f, 'v) term \<Rightarrow> ('f, 'v) rseq \<Rightarrow> ('f, 'v) step list"
where
  "rseq_to_step_list s [] = []"
| "rseq_to_step_list s ((p,lr,t) # steps) = (
     let u = subt_at s p;
         v = subt_at t p;
         \<sigma> = (case match_list Var [(fst lr, u), (snd lr, v)] of
           Some \<tau> \<Rightarrow> \<tau>
         | None \<Rightarrow> undefined)
      in (s,lr,p,\<sigma>,True,t) # rseq_to_step_list t steps)"

primrec rseq_to_ddseq :: "('f, 'v) term \<times> ('f, 'v) rseq \<Rightarrow> ('f, 'v) seq"
  where "rseq_to_ddseq (s, rseq) = (s, rseq_to_step_list s rseq)"

lemma check_rsteps'_sound_dd_seq:
  assumes "isOK (check_rsteps' R rseq s w)"
  shows "rseq_to_ddseq (s, rseq) \<in> seq (set R) \<and> DD2_last (rseq_to_ddseq (s, rseq)) = w"
using assms
proof (induct rseq arbitrary: s)
  case Nil
  have "rseq_to_ddseq (s, []) \<in> seq (set R)" using seq.intros by auto
  moreover have "DD2_last (rseq_to_ddseq (s, [])) = s" by auto
  moreover have "s = w" using Nil check_rsteps_sound by fastforce
  ultimately show ?case by auto
next
  case (Cons r rs)
  show ?case 
  proof (cases r)
    case (fields p lr t)
    with Cons have IH: "rseq_to_ddseq (t, rs) \<in> seq (set R)" and 
      "DD2_last (rseq_to_ddseq (t, rs)) = w" by auto
    with fields have last:"DD2_last (rseq_to_ddseq (s, r # rs)) = w" 
      by (simp add: get_target_def)
    let ?\<sigma> = "case match_list Var [(fst lr, s |_ p), (snd lr, t |_ p)] of Some \<tau> \<Rightarrow> \<tau>"
    from fields Cons(2) obtain \<tau> where 
      "lr \<in> set R" and "p \<in> poss s" and "p \<in> poss t" and 
      "ctxt_of_pos_term p s = ctxt_of_pos_term p t" and
      "match_list Var [(fst lr, s |_ p), (snd lr, t |_ p)] = Some \<tau>" 
      apply (simp del: check_rstep')
      unfolding check_rstep'_def unfolding check_prop_rstep'_def 
      unfolding check_prop_rstep_rule_def unfolding Let_def 
      apply (cases "match_list Var [(fst lr, s |_ p), (snd lr, t |_ p)]") by auto
    moreover from this  match_list_matches have "(fst lr) \<cdot> ?\<sigma> = s |_ p \<and> (snd lr) \<cdot> ?\<sigma> = t |_ p" 
      by fastforce
    ultimately have "(s, t) \<in> rstep_r_p_s (set R) lr p ?\<sigma>" 
      unfolding rstep_r_p_s_def' by (cases lr, auto simp: replace_at_ident)
    with IH have "rseq_to_ddseq (s, r # rs) \<in> seq (set R)" using fields seq.intros by fastforce
    with last show ?thesis by blast
  qed
qed

fun eseq_to_step_list :: "('f, 'v) term \<Rightarrow> ('f, 'v) eseq \<Rightarrow> ('f, 'v) step list"
where
  "eseq_to_step_list s [] = []"
| "eseq_to_step_list s ((p,lr,b,t) # steps) = (
     let u = subt_at s p; v = subt_at t p in
     if b then (let \<sigma> = (case match_list Var [(fst lr, u), (snd lr, v)] of
           Some \<tau> \<Rightarrow> \<tau>
         | None \<Rightarrow> undefined)
      in (s,lr,p,\<sigma>,b,t) # eseq_to_step_list t steps)
     else (let \<sigma> = (case match_list Var [(snd lr, u), (fst lr, v)] of
           Some \<tau> \<Rightarrow> \<tau>
         | None \<Rightarrow> undefined)
      in (t,lr,p,\<sigma>,b,s) # eseq_to_step_list t steps))"

primrec eseq_to_ddconv :: "('f, 'v) term \<times> ('f, 'v) eseq \<Rightarrow> ('f, 'v) conv"
  where "eseq_to_ddconv (s, eseq) = (s, eseq_to_step_list s eseq)"

(* checking conversions without allowing variants of rules *)
definition
  check_estep' ::
    "('f:: showl, 'v:: showl) equation list \<Rightarrow>
    pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> bool \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow>
    showsl check"
where
  "check_estep' E p rule l_to_r s t \<equiv> do {
    check (rule \<in> set E)
      (showsl_rule rule \<circ> showsl_lit (STR ''is not an equation of'') \<circ> showsl_nl \<circ> showsl_eqs E \<circ> showsl_nl);
    check (p \<in> poss s) (showsl p \<circ> showsl_lit (STR '' is not a position of '') \<circ> showsl s \<circ> showsl_nl);
    check (p \<in> poss t) (showsl p \<circ> showsl_lit (STR '' is not a position of '') \<circ> showsl t \<circ> showsl_nl);
    let C = ctxt_of_pos_term p s;
    let D = ctxt_of_pos_term p t;
    let u = subt_at s p;
    let v = subt_at t p;
    let rrule = (if l_to_r then rule else (snd rule, fst rule));
    let err = (showsl_lit (STR ''the term '') \<circ> showsl t
      \<circ> showsl_lit (STR '' does not result from a proper application of term '') \<circ> showsl s \<circ> showsl_lit (STR '' using equation '')
      \<circ> showsl_nl \<circ> showsl_eq rrule \<circ> showsl_lit (STR '' at position '') \<circ> showsl p \<circ> showsl_nl);
    (case match_list Var [(fst rrule, u), (snd rrule, v)] of
      Some \<sigma> \<Rightarrow> check (C = D) err
    | None \<Rightarrow> error err)
  }"

fun
  check_conversion' ::
    "('f:: showl, 'v:: showl) equation list \<Rightarrow> 
     ('f, 'v) eseq \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> showsl check"
where
  "check_conversion' E [] s u = check (s = u) (
    showsl_lit (STR ''the last term of the conversion '') \<circ> showsl_nl \<circ> showsl s \<circ> showsl_nl \<circ>
    showsl_lit (STR ''does not correspond to the goal term'') \<circ> showsl_nl \<circ> showsl u \<circ> showsl_nl)"
| "check_conversion' E ((p, r, l_to_r, t) # c) s u = do {
    check_estep' E p r l_to_r s t;
    check_conversion' E c t u
  }"

definition eseq_last :: "('f,'v)term \<Rightarrow> ('f,'v) eseq \<Rightarrow> ('f,'v)term"
  where "eseq_last s steps \<equiv> last (s # map (\<lambda> (_,_,_,s). s) steps)"

lemma check_conversion_sound_dd_conv:
  assumes "isOK (check_conversion' E eseq s w)"
  shows "eseq_to_ddconv (s, eseq) \<in> conv (set E) \<and> DD2_last (eseq_to_ddconv (s, eseq)) = w"
using assms
proof (induct eseq arbitrary: s)
  case Nil
  have "eseq_to_ddconv (s, []) \<in> conv (set E)"  by (auto intro: conv.intros)
  moreover have "DD2_last (eseq_to_ddconv (s, [])) = s" by auto
  moreover have "s = w" using Nil by fastforce
  ultimately show ?case by auto
next
  case (Cons r rs)
  show ?case 
  proof (cases r)
    case (fields p lr b t)
    with Cons have IH: "eseq_to_ddconv (t, rs) \<in> conv (set E)" and
      "DD2_last (eseq_to_ddconv (t, rs)) = w" by auto
    with fields have last:"DD2_last (eseq_to_ddconv (s, r # rs)) = w"
      by (cases b) (auto simp: get_target_def)
    show ?thesis
    proof (cases b)
      case True
      let ?\<sigma> = "case match_list Var [(fst lr, s |_ p), (snd lr, t |_ p)] of Some \<tau> \<Rightarrow> \<tau>"
      from fields Cons(2) True obtain \<tau> where
        "lr \<in> set E" and "p \<in> poss s" and "p \<in> poss t" and
        "ctxt_of_pos_term p s = ctxt_of_pos_term p t" and
        "match_list Var [(fst lr, s |_ p), (snd lr, t |_ p)] = Some \<tau>"
        apply (simp add: check_estep'_def Let_def)
        apply (cases "match_list Var [(fst lr, s |_ p), (snd lr, t |_ p)]") by auto
      moreover with match_list_matches have "(fst lr) \<cdot> ?\<sigma> = s |_ p \<and> (snd lr) \<cdot> ?\<sigma> = t |_ p"
        by fastforce
      ultimately have step:"(s, t) \<in> rstep_r_p_s (set E) lr p ?\<sigma>"
        unfolding rstep_r_p_s_def' by (cases lr, auto simp: replace_at_ident)
      from IH have "eseq_to_ddconv (s, r # rs) \<in> conv (set E)" unfolding fields
        using True conv.intros(2)[OF step] by simp
      with last show ?thesis by blast
    next
      case False
      let ?\<sigma> = "case match_list Var [(snd lr, s |_ p), (fst lr, t |_ p)] of Some \<tau> \<Rightarrow> \<tau>"
      from fields Cons(2) False obtain \<tau> where
        "lr \<in> set E" and "p \<in> poss s" and "p \<in> poss t" and
        "ctxt_of_pos_term p t = ctxt_of_pos_term p s" and
        "match_list Var [(snd lr, s |_ p), (fst lr, t |_ p)] = Some \<tau>"
        apply (simp add: check_estep'_def Let_def)
        apply (cases "match_list Var [(snd lr, s |_ p), (fst lr, t |_ p)]") by auto
      moreover with match_list_matches have "(fst lr) \<cdot> ?\<sigma> = t |_ p \<and> (snd lr) \<cdot> ?\<sigma> = s |_ p"
        by fastforce
      ultimately have step:"(t, s) \<in> rstep_r_p_s (set E) lr p ?\<sigma>"
        unfolding rstep_r_p_s_def' by (cases lr, auto simp: replace_at_ident)
      from IH have "eseq_to_ddconv (s, r # rs) \<in> conv (set E)" unfolding fields
        using False conv.intros(3)[OF step] by simp
      with last show ?thesis by blast
    qed
  qed
qed

definition check_ELD_1_nat :: "nat \<Rightarrow> nat \<Rightarrow> nat list \<Rightarrow> nat list \<Rightarrow> nat list \<Rightarrow> showsl check"
where
  "check_ELD_1_nat \<beta> \<alpha> \<sigma>\<^sub>1 \<sigma>\<^sub>2 \<sigma>\<^sub>3 = do {
     check_allm (\<lambda>x. check (x < \<beta>) 
       (showsl_lit (STR ''the labels are not decreasing: '') \<circ> showsl x \<circ> showsl_lit (STR '' is not smaller '') \<circ> showsl \<beta> \<circ> showsl_nl)) \<sigma>\<^sub>1;
     check_allm (\<lambda>x. check (x \<le> \<alpha>)
       (showsl_lit (STR ''the labels are not decreasing: '') \<circ> showsl x \<circ> showsl_lit (STR '' is not smaller equal '') \<circ> showsl \<alpha> \<circ> showsl_nl)) \<sigma>\<^sub>2;
     check (length \<sigma>\<^sub>2 \<le> 1) (showsl_lit (STR ''  the length of the middle sequence is greater 1\<newline>''));
     check_allm (\<lambda>x. check (x < \<alpha> \<or> x < \<beta>)
       (showsl_lit (STR ''the labels are not decreasing: '') \<circ> showsl x \<circ> showsl_lit (STR '' is not smaller '') 
         \<circ> showsl \<alpha> \<circ> showsl_lit (STR '' or smaller '') \<circ> showsl \<beta> \<circ> showsl_nl)) \<sigma>\<^sub>3
   }"

lemma check_ELD_1_nat :
  assumes "isOK (check_ELD_1_nat \<beta> \<alpha> \<sigma>\<^sub>1 \<sigma>\<^sub>2 \<sigma>\<^sub>3)"
  shows "ELD_1 ({(n, m). n < m},{(n, m). n \<le> m}) \<beta> \<alpha> \<sigma>\<^sub>1 \<sigma>\<^sub>2 \<sigma>\<^sub>3"
using assms
unfolding check_ELD_1_nat_def ELD_1_def ds_def
by auto

definition split_seq :: "nat \<Rightarrow> nat \<Rightarrow> nat list \<Rightarrow> nat list \<times> nat list \<times> nat list"
where
  "split_seq \<alpha> \<beta> ss = (
    let (ss', r) = span (\<lambda>n. n < \<alpha>) ss in
      case r of
        Nil \<Rightarrow> (ss',[],[])
      | Cons h t \<Rightarrow> if h \<le> \<beta> then (ss',[h],t) else (ss',[],r)
  )"

lemma split_seq_is_split:
  assumes "split_seq \<alpha> \<beta> ss = (ss\<^sub>1, ss\<^sub>2, ss\<^sub>3)"
  shows "ss\<^sub>1 @ ss\<^sub>2 @ ss\<^sub>3 = ss"
using assms 
by (auto split: list.splits if_splits simp: split_seq_def) (metis takeWhile_dropWhile_id)


lemma split_complete:
  assumes "split_seq \<alpha> \<beta> ss = (ss\<^sub>1, ss\<^sub>2, ss\<^sub>3)"
  and "\<forall> x \<in> set ss\<^sub>3. x < \<alpha> \<or> x < \<beta>"
  shows "ELD_1 ({(n, m). n < m},{(n, m). n \<le> m}) \<alpha> \<beta> ss\<^sub>1 ss\<^sub>2 ss\<^sub>3"
proof -
  obtain ss' r where sp: "(ss', r) = span (\<lambda>n. n < \<alpha>) ss" by auto
  show ?thesis
  proof (cases r)
    case Nil
    with assms sp have "ss\<^sub>1 = ss'" "ss\<^sub>2 = []" "ss\<^sub>3 = []"
      unfolding split_seq_def Let_def by auto
    with sp show ?thesis unfolding ELD_1_def using set_takeWhileD under_def by fastforce
  next
    case (Cons h t)
    show ?thesis
    proof(cases "h \<le> \<beta>")
      case True
      with assms sp Cons have "ss\<^sub>1 = ss'" "ss\<^sub>2 = [h]" "ss\<^sub>3 = t"
        unfolding split_seq_def Let_def by (case_tac "dropWhile (\<lambda>n. n < \<alpha>) ss", auto)+
      with sp True Cons show ?thesis unfolding ELD_1_def ds_def apply auto
        by (metis set_takeWhileD) (metis assms(2))
    next
      case False
      with assms sp Cons have "ss\<^sub>1 = ss'" "ss\<^sub>2 = []" "ss\<^sub>3 = r"
        unfolding split_seq_def Let_def by (case_tac "dropWhile (\<lambda>n. n < \<alpha>) ss", auto)+
      with sp False Cons show ?thesis unfolding ELD_1_def ds_def apply auto
        by (metis set_takeWhileD) (metis assms(2))
    qed
  qed
qed

definition check_cpeak_eld :: 
  "('f :: showl, 'v :: showl) rules \<Rightarrow> ('f, 'v) lpeak \<Rightarrow> ('f, 'v) rule \<Rightarrow>('f, 'v) rseq  \<Rightarrow> ('f, 'v) rseq  \<Rightarrow> 
  (('f, 'v) rule \<Rightarrow> nat) \<Rightarrow> showsl check"
where
  "check_cpeak_eld R p cp j1 j2 lab = 
     (case p of ((s1,r1,p1,\<sigma>1,True,t1),(s2,r2,p2,\<sigma>2,True,t2)) \<Rightarrow>
      do {
        check (instance_rule (t1, t2) cp) id;
        let u = rseq_last (fst cp) j1;
        let v = rseq_last (snd cp) j2;
        check_rsteps' R j1 (fst cp) u;
        check_rsteps' R j2 (snd cp) v;
        check (u = v) (showsl_lit (STR ''the rewrite sequences end in different terms: '') 
          \<circ> showsl u \<circ> showsl_lit (STR '' and '') \<circ> showsl v \<circ> showsl_nl);
        let \<alpha> = lab r1;
        let \<beta> = lab r2;
        let \<tau> = map (rule_labeling lab) (snd (rseq_to_ddseq (t1, j1)));
        let \<sigma> = map (rule_labeling lab) (snd (rseq_to_ddseq (t2, j2)));
        let (\<tau>\<^sub>1, \<tau>\<^sub>2, \<tau>\<^sub>3) = split_seq \<alpha> \<beta> \<tau>;
        let (\<sigma>\<^sub>1, \<sigma>\<^sub>2, \<sigma>\<^sub>3) = split_seq \<beta> \<alpha> \<sigma>;
        check_ELD_1_nat \<alpha> \<beta> \<tau>\<^sub>1 \<tau>\<^sub>2 \<tau>\<^sub>3;
        check_ELD_1_nat \<beta> \<alpha> \<sigma>\<^sub>1 \<sigma>\<^sub>2 \<sigma>\<^sub>3        
      })"
  
lemma check_cpeak_eld:
  defines "r \<equiv> ({(n, m) . n < m}, {(n, m) . n \<le> m})"
  assumes "(b, (s1,r1,p1,\<sigma>1,True,t1),(s2,r2,p2,\<sigma>2,True,t2)) \<in> critical_peaks ren (set R)" (is "(b, ?p) \<in> critical_peaks _ (set R)")
  and "isOK (check_cpeak_eld R ((s1,r1,p1,\<sigma>1,True,t1),(s2,r2,p2,\<sigma>2,True,t2)) cp j1 j2 lab)"
  shows "eld (set R) (rule_labeling lab) r ?p" (is "eld _ ?l _ _")
proof -
  let ?u = "rseq_last (fst cp) j1"
  let ?v = "rseq_last (snd cp) j2"
  let ?j1cp = "rseq_to_ddseq ((fst cp), j1)"
  let ?j2cp = "rseq_to_ddseq ((snd cp), j2)"
  let ?j1t = "rseq_to_ddseq (t1, j1)"
  let ?j2t = "rseq_to_ddseq (t2, j2)"
  from assms have 
    steps1: "isOK (check_rsteps' R j1 (fst cp) ?u)" and 
    steps2: "isOK (check_rsteps' R j2 (snd cp) ?v)"and 
    eq: "?u = ?v" and
    inst: "instance_rule (t1, t2) cp"
    by (auto simp add: Let_def check_cpeak_eld_def)
  from inst obtain \<rho> where \<rho>: "t1 = fst cp \<cdot> \<rho> \<and> t2 = snd cp \<cdot> \<rho>" 
    unfolding instance_rule_def by auto
  let ?j1\<rho> = "seq_close ?j1cp \<box> \<rho>"
  let ?j2\<rho> = "seq_close ?j2cp \<box> \<rho>"
  from assms obtain \<sigma>1 \<sigma>2 \<sigma>3 \<tau>1 \<tau>2 \<tau>3  where 
    \<sigma>: "(\<sigma>1,\<sigma>2,\<sigma>3) = split_seq (lab r2) (lab r1) (map ?l (snd ?j2t))" and
    \<tau>: "(\<tau>1,\<tau>2,\<tau>3) = split_seq (lab r1) (lab r2) (map ?l (snd ?j1t))" and
    \<sigma>ok: "isOK (check_ELD_1_nat (lab r2) (lab r1) \<sigma>1 \<sigma>2 \<sigma>3)" and
    \<tau>ok: "isOK (check_ELD_1_nat (lab r1) (lab r2) \<tau>1 \<tau>2 \<tau>3)" 
    unfolding Let_def check_cpeak_eld_def using check_ELD_1_nat by auto
  moreover have "map ?l (snd ?j1t) = map ?l (snd ?j1\<rho>)" by (induct j1, auto)
  moreover have "map ?l (snd ?j2t) = map ?l (snd ?j2\<rho>)" by (induct j2, auto)
  ultimately have \<sigma>2: "split_seq (lab r2) (lab r1) (map ?l (snd ?j2\<rho>)) = (\<sigma>1,\<sigma>2,\<sigma>3)" 
   and \<tau>2: "split_seq (lab r1) (lab r2) (map ?l (snd ?j1\<rho>)) = (\<tau>1,\<tau>2,\<tau>3)" by auto
  from assms(2) critical_peak_is_local_peak have lp:"?p \<in> local_peaks (set R)" by fast
  from steps1 check_rsteps'_sound_dd_seq have j1cpseq: "?j1cp \<in> seq (set R)" and 
    j1cplast:"DD2_last ?j1cp = ?u" by auto
  from steps2 check_rsteps'_sound_dd_seq have j2cpseq: "?j2cp \<in> seq (set R)" and 
    j2cplast:"DD2_last ?j2cp = ?v" by auto
  from seq_super_closed1[OF j1cpseq, of "\<box>" \<rho>] \<rho> j1cplast have j1seq:"?j1\<rho> \<in> seq (set R)" 
    and " get_target (fst ?p) = first ?j1\<rho>" and "DD2_last ?j1\<rho> = ?u \<cdot> \<rho>"
    unfolding get_target_def by auto
  moreover from seq_super_closed1[OF j2cpseq, of "\<box>" \<rho>] \<rho> j2cplast have j2seq:"?j2\<rho> \<in> seq (set R)" 
    and " get_target (snd ?p) = first ?j2\<rho>" and "DD2_last ?j2\<rho> = ?v \<cdot> \<rho>"
    unfolding get_target_def by auto
  moreover from seq_dec[OF j2seq split_seq_is_split[OF \<sigma>2,symmetric]] obtain jr1 jr23 where
    "?j2\<rho> = seq_concat jr1 jr23" and "jr1 \<in> seq (set R)" and "jr23 \<in> seq (set R)" and
    "map ?l (snd jr1) = \<sigma>1" and \<sigma>23:"map ?l (snd jr23) = \<sigma>2 @ \<sigma>3" and "first ?j2\<rho> = first jr1" and
    "DD2_last jr1 = first jr23" and "DD2_last jr23 = DD2_last ?j2\<rho>" by blast
  moreover from seq_dec[OF \<open>jr23 \<in> seq (set R)\<close> \<sigma>23] obtain jr2 jr3 where
    "jr23 = seq_concat jr2 jr3" and "jr2 \<in> seq (set R)" and "jr3 \<in> seq (set R)" and
    "map ?l (snd jr2) = \<sigma>2" and "map ?l (snd jr3) = \<sigma>3" and "first jr23 = first jr2" and 
    "DD2_last jr2 = first jr3" and "DD2_last jr3 = DD2_last jr23" by blast
  moreover from seq_dec[OF j1seq split_seq_is_split[OF \<tau>2,symmetric]] obtain jl1 jl23 where
    "?j1\<rho> = seq_concat jl1 jl23" and "jl1 \<in> seq (set R)" and "jl23 \<in> seq (set R)" and
    "map ?l (snd jl1) = \<tau>1" and \<tau>23:"map ?l (snd jl23) = \<tau>2 @ \<tau>3" and "first ?j1\<rho> = first jl1" and
    "DD2_last jl1 = first jl23" and "DD2_last jl23 = DD2_last ?j1\<rho>" by blast
  moreover from seq_dec[OF \<open>jl23 \<in> seq (set R)\<close> \<tau>23] obtain jl2 jl3 where
    "jl23 = seq_concat jl2 jl3" and "jl2 \<in> seq (set R)" and "jl3 \<in> seq (set R)" and
    "map ?l (snd jl2) = \<tau>2" and "map ?l (snd jl3) = \<tau>3" and "first jl23 = first jl2" and 
    "DD2_last jl2 = first jl3" and "DD2_last jl3 = DD2_last jl23" by blast
  ultimately moreover have ld:"ld_trs (set R) ?p jl1 jl2 jl3 jr1 jr2 jr3"
    unfolding ld_trs_def using lp eq by (auto simp: seq_concat_def) 
  ultimately have "eld_conv (rule_labeling lab) r ?p jl1 jl2 jl3 jr1 jr2 jr3"
    unfolding eld_conv_def r_def using check_ELD_1_nat[OF \<sigma>ok] check_ELD_1_nat[OF \<tau>ok]
    by auto
  with ld show ?thesis unfolding eld_def by fast
qed

definition "cpeak_instance s cp s' cp' = 
  (match_list (\<lambda> _. s) [(s, s'), (fst cp, fst cp'), (snd cp, snd cp')] \<noteq> None)"

definition check_cpeak_eldc :: 
  "('f :: showl, 'v :: showl) rules \<Rightarrow> ('f, 'v) lpeak \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) rule \<Rightarrow> 
   ('f, 'v) eseq  \<Rightarrow> ('f, 'v) rseq  \<Rightarrow> ('f, 'v) eseq  \<Rightarrow> ('f, 'v) eseq  \<Rightarrow> ('f, 'v) rseq  \<Rightarrow> ('f, 'v) eseq  \<Rightarrow>
  (('f, 'v) rule \<Rightarrow> nat) \<Rightarrow> nat option \<Rightarrow> showsl check"
where
  "check_cpeak_eldc R p s cp cl\<^sub>1 sl cl\<^sub>2 cr\<^sub>1 sr cr\<^sub>2 lab n = 
     (case p of ((s1,r1,p1,\<sigma>1,True,t1),(s2,r2,p2,\<sigma>2,True,t2)) \<Rightarrow>
      do {
        check (cpeak_instance  s cp s1 (t1, t2)) id;
        let u\<^sub>1 = eseq_last (fst cp) cl\<^sub>1;
        let v\<^sub>1 = eseq_last (snd cp) cr\<^sub>1;
        let u\<^sub>2 = rseq_last u\<^sub>1 sl;
        let v\<^sub>2 = rseq_last v\<^sub>1 sr;
        let u\<^sub>3 = eseq_last u\<^sub>2 cl\<^sub>2;
        let v\<^sub>3 = eseq_last v\<^sub>2 cr\<^sub>2;
        check_conversion' R cl\<^sub>1 (fst cp) u\<^sub>1;
        check_rsteps' R sl u\<^sub>1 u\<^sub>2;
        check_conversion' R cl\<^sub>2 u\<^sub>2 u\<^sub>3;
        check_conversion' R cr\<^sub>1 (snd cp) v\<^sub>1;
        check_rsteps' R sr v\<^sub>1 v\<^sub>2;
        check_conversion' R cr\<^sub>2 v\<^sub>2 v\<^sub>3;
        check (u\<^sub>3 = v\<^sub>3) (showsl_lit (STR ''the conversions end in different terms: '') \<circ> 
          showsl u\<^sub>3 \<circ> showsl_lit (STR '' and '') \<circ> showsl v\<^sub>3 \<circ> showsl_nl);
        let \<alpha> = lab r1;
        let \<beta> = lab r2;
        let cl\<^sub>1 = snd (eseq_to_ddconv (fst cp, cl\<^sub>1));
        let sl = snd (rseq_to_ddseq (u\<^sub>1, sl));
        let cl\<^sub>2 = snd (eseq_to_ddconv (u\<^sub>2, cl\<^sub>2));
        let cr\<^sub>1 = snd (eseq_to_ddconv (snd cp, cr\<^sub>1));
        let sr = snd (rseq_to_ddseq (v\<^sub>1, sr));
        let cr\<^sub>2 = snd (eseq_to_ddconv (v\<^sub>2, cr\<^sub>2));
        let \<tau>\<^sub>1 = map (rule_labeling lab) cl\<^sub>1; 
        let \<tau>\<^sub>2 = map (rule_labeling lab) sl;
        let \<tau>\<^sub>3 = map (rule_labeling lab) cl\<^sub>2;
        let \<sigma>\<^sub>1 = map (rule_labeling lab) cr\<^sub>1;
        let \<sigma>\<^sub>2 = map (rule_labeling lab) sr;
        let \<sigma>\<^sub>3 = map (rule_labeling lab) cr\<^sub>2;
        check_ELD_1_nat \<alpha> \<beta> \<tau>\<^sub>1 \<tau>\<^sub>2 \<tau>\<^sub>3;
        check_ELD_1_nat \<beta> \<alpha> \<sigma>\<^sub>1 \<sigma>\<^sub>2 \<sigma>\<^sub>3;
        case n of 
          Some n \<Rightarrow> do {
            let check_reachable = (\<lambda>t. check (List.member (reachable_terms R s n) t)
              (showsl_lit (STR ''the fan property is violated: '') \<circ> showsl t \<circ> 
               showsl_lit (STR '' is not reachable from '') \<circ> showsl s \<circ> 
               showsl_lit (STR '' in '') \<circ> showsl n \<circ> showsl_lit (STR '' steps\<newline>''))); 
            check_allm check_reachable (map get_source cl\<^sub>1);
            check_allm check_reachable (map get_source sl);
            check_allm check_reachable (map get_source cl\<^sub>2);
            check_allm check_reachable (map get_source cr\<^sub>1);
            check_allm check_reachable (map get_source sr);
            check_allm check_reachable (map get_source cr\<^sub>2)
          }
        | None \<Rightarrow> succeed
      })"

lemma get_source_closed:
  assumes "i < length ss"
  shows "get_source ((map (\<lambda>step. step_close step C \<sigma>) ss) ! i) = C\<langle>get_source (ss ! i) \<cdot> \<sigma>\<rangle>"
using assms
proof (induct ss arbitrary: i)
  case Nil
  then show ?case by (auto simp: get_source_def)
next
  case (Cons step ss)
  then show ?case by (cases step, cases i) (auto simp: get_source_def) 
qed

lemma check_cpeak_eldc:
  defines "r \<equiv> ({(n, m) . n < m}, {(n, m) . n \<le> m})"
  assumes "(b, (s1,r1,p1,\<sigma>1,True,t1),(s2,r2,p2,\<sigma>2,True,t2)) \<in> critical_peaks ren (set R)" (is "(b, ?p) \<in> critical_peaks ren (set R)")
  and "isOK (check_cpeak_eldc R ((s1,r1,p1,\<sigma>1,True,t1),(s2,r2,p2,\<sigma>2,True,t2)) s cp cl\<^sub>1 sl cl\<^sub>2 cr\<^sub>1 sr cr\<^sub>2 lab n)"
  and "n = None \<longrightarrow> linear_trs (set R)"
  shows "eld_fan (set R) (rule_labeling lab) r ?p \<or> eldc (set R) (rule_labeling lab) r ?p \<and> linear_trs (set R)" (is "eld_fan _ ?l _ _ \<or> _")
proof -
  let ?u\<^sub>1 = "eseq_last (fst cp) cl\<^sub>1"
  let ?v\<^sub>1 = "eseq_last (snd cp) cr\<^sub>1"
  let ?u\<^sub>2 = "rseq_last ?u\<^sub>1 sl"
  let ?v\<^sub>2 = "rseq_last ?v\<^sub>1 sr"
  let ?u\<^sub>3 = "eseq_last ?u\<^sub>2 cl\<^sub>2"
  let ?v\<^sub>3 = "eseq_last ?v\<^sub>2 cr\<^sub>2"
  let ?cl\<^sub>1 = "eseq_to_ddconv (fst cp, cl\<^sub>1)"
  let ?sl = "rseq_to_ddseq (?u\<^sub>1, sl)"
  let ?cl\<^sub>2 = "eseq_to_ddconv (?u\<^sub>2, cl\<^sub>2)"
  let ?cr\<^sub>1 = "eseq_to_ddconv (snd cp, cr\<^sub>1)"
  let ?sr = "rseq_to_ddseq (?v\<^sub>1, sr)"
  let ?cr\<^sub>2 = "eseq_to_ddconv (?v\<^sub>2, cr\<^sub>2)"
  from assms have 
    cl1: "isOK (check_conversion' R cl\<^sub>1 (fst cp) ?u\<^sub>1)" and
    sl: "isOK (check_rsteps' R sl ?u\<^sub>1 ?u\<^sub>2)" and
    cl2: "isOK (check_conversion' R cl\<^sub>2 ?u\<^sub>2 ?u\<^sub>3)" and
    cr1: "isOK (check_conversion' R cr\<^sub>1 (snd cp) ?v\<^sub>1)" and
    sr: "isOK (check_rsteps' R sr ?v\<^sub>1 ?v\<^sub>2)" and
    cr2: "isOK (check_conversion' R cr\<^sub>2 ?v\<^sub>2 ?v\<^sub>3)" and 
    eq: "?u\<^sub>3 = ?v\<^sub>3" and inst: "cpeak_instance s cp s1 (t1, t2)"
    by (auto simp add: Let_def check_cpeak_eldc_def)
  from inst obtain \<rho> where "match_list (\<lambda>_. s) [(s, s1), (fst cp, t1), (snd cp, t2)] = Some \<rho>" 
    unfolding cpeak_instance_def by auto
  from match_list_sound[OF this] have \<rho>: "t1 = fst cp \<cdot> \<rho> \<and> t2 = snd cp \<cdot> \<rho> \<and> s1 = s \<cdot> \<rho>" 
    unfolding matchers_def by auto
  let ?s = "\<lambda>s. seq_close s \<box> \<rho>"
  let ?c = "\<lambda>c. conv_close c \<box> \<rho>"
  have "ldc_trs (set R) ?p (?c ?cl\<^sub>1) (?s ?sl) (?c ?cl\<^sub>2) (?c ?cr\<^sub>1) (?s ?sr) (?c ?cr\<^sub>2)"
    unfolding ldc_trs_def
  proof (intro conjI)
    from assms(2) critical_peak_is_local_peak show lp:"?p \<in> local_peaks (set R)" by fast
    from check_conversion_sound_dd_conv[OF cl1] have "?cl\<^sub>1 \<in> conv (set R)" and "DD2_last ?cl\<^sub>1 = ?u\<^sub>1" by auto
    with conv_super_closed1[OF this(1), of \<box> \<rho>] show "(?c ?cl\<^sub>1) \<in> conv (set R)" and "DD2_last (?c ?cl\<^sub>1) = first (?s ?sl)" by auto
    from check_rsteps'_sound_dd_seq[OF sl] have "?sl \<in> seq (set R)" and "DD2_last ?sl = ?u\<^sub>2" by auto
    with seq_super_closed1[OF this(1), of \<box> \<rho>] show "(?s ?sl) \<in> seq (set R)" and "DD2_last (?s ?sl) = first (?c ?cl\<^sub>2)" by auto
    from check_conversion_sound_dd_conv[OF cl2] have "?cl\<^sub>2 \<in> conv (set R)" and "DD2_last ?cl\<^sub>2 = ?u\<^sub>3" by auto
    with conv_super_closed1[OF this(1), of \<box> \<rho>] have "(?c ?cl\<^sub>2) \<in> conv (set R)" and u3:"DD2_last (?c ?cl\<^sub>2) = ?u\<^sub>3 \<cdot> \<rho>" by auto
    from check_conversion_sound_dd_conv[OF cr1] have "?cr\<^sub>1 \<in> conv (set R)" and "DD2_last ?cr\<^sub>1 = ?v\<^sub>1" by auto
    with conv_super_closed1[OF this(1), of \<box> \<rho>] show "(?c ?cr\<^sub>1) \<in> conv (set R)" and "DD2_last (?c ?cr\<^sub>1) = first (?s ?sr)" by auto
    from check_rsteps'_sound_dd_seq[OF sr] have "?sr \<in> seq (set R)" and "DD2_last ?sr = ?v\<^sub>2" by auto
    with seq_super_closed1[OF this(1), of \<box> \<rho>] show "(?s ?sr) \<in> seq (set R)" and "DD2_last (?s ?sr) = first (?c ?cr\<^sub>2)" by auto
    from check_conversion_sound_dd_conv[OF cr2] have "?cr\<^sub>2 \<in> conv (set R)" and v3:"DD2_last ?cr\<^sub>2 = ?v\<^sub>3" by auto
    with conv_super_closed1[OF this(1), of \<box> \<rho>] have "(?c ?cr\<^sub>2) \<in> conv (set R)" and v3:"DD2_last (?c ?cr\<^sub>2) = ?v\<^sub>3 \<cdot> \<rho>" by auto
    show "get_target (fst ?p) = first (?c ?cl\<^sub>1)" unfolding get_target_def using \<rho> by auto
    show "get_target (snd ?p) = first (?c ?cr\<^sub>1)" unfolding get_target_def using \<rho> by auto
    from u3 v3 eq show "DD2_last (?c ?cl\<^sub>2) = DD2_last (?c ?cr\<^sub>2)" by auto
    show "(?c ?cl\<^sub>2) \<in> conv (set R)" and "(?c ?cr\<^sub>2) \<in> conv (set R)" by fact+
  qed
  moreover have "eld_conv ?l r ?p (?c ?cl\<^sub>1) (?s ?sl) (?c ?cl\<^sub>2) (?c ?cr\<^sub>1) (?s ?sr) (?c ?cr\<^sub>2)"
  proof -
    from assms[unfolded check_cpeak_eldc_def Let_def]
    have "ELD_1 r (?l (fst ?p)) (?l (snd ?p)) (map ?l (snd ?cl\<^sub>1)) (map ?l (snd ?sl)) (map ?l (snd ?cl\<^sub>2))"
      and "ELD_1 r (?l (snd ?p)) (?l (fst ?p)) (map ?l (snd ?cr\<^sub>1)) (map ?l (snd ?sr)) (map ?l (snd ?cr\<^sub>2))"
      using check_ELD_1_nat by auto
    moreover have "map ?l (snd ?cl\<^sub>1) = map ?l (snd (?c ?cl\<^sub>1))" and
      "map ?l (snd ?sl) = map ?l (snd (?s ?sl))" and
      "map ?l (snd ?cl\<^sub>2) = map ?l (snd (?c ?cl\<^sub>2))" and 
      "map ?l (snd ?cr\<^sub>1) = map ?l (snd (?c ?cr\<^sub>1))" and
      "map ?l (snd ?sr) = map ?l (snd (?s ?sr))" and
      "map ?l (snd ?cr\<^sub>2) = map ?l (snd (?c ?cr\<^sub>2))"by auto
    ultimately show ?thesis unfolding eld_conv_def by metis
  qed
  moreover have "fan (set R) ?p (?c ?cl\<^sub>1) (?s ?sl) (?c ?cl\<^sub>2) (?c ?cr\<^sub>1) (?s ?sr) (?c ?cr\<^sub>2) \<or> linear_trs (set R)"
  proof (cases n)
    case (Some n)
    show ?thesis unfolding fan_def
    proof (intro disjI1 conjI allI impI)
      fix i assume i:"i < length (snd (?c ?cl\<^sub>1))"
      then have i2:"i < length (snd ?cl\<^sub>1)" by auto
      then have "snd ?cl\<^sub>1 ! i \<in> set (snd ?cl\<^sub>1)" using nth_mem by auto
      with assms[unfolded Let_def check_cpeak_eldc_def] Some
      have "List.member (reachable_terms R s n) (get_source (snd ?cl\<^sub>1 ! i))" by auto
      then have "(s, get_source (snd ?cl\<^sub>1 ! i)) \<in> (rstep (set R))\<^sup>*"
        using reachable_terms in_set_member by metis
      then show "(get_source (fst ?p), get_source (snd (?c ?cl\<^sub>1) ! i)) \<in> (rstep (set R))\<^sup>*" 
        using get_source_closed[OF i2, of \<box> \<rho>] \<rho>
        by (auto intro: rsteps_closed_subst simp: get_source_def)
    next
      fix i assume i:"i < length (snd (?c ?cl\<^sub>2))"
      then have i2:"i < length (snd ?cl\<^sub>2)" by auto
      then have "snd ?cl\<^sub>2 ! i \<in> set (snd ?cl\<^sub>2)" using nth_mem by auto
      with assms[unfolded Let_def check_cpeak_eldc_def] Some
      have "List.member (reachable_terms R s n) (get_source (snd ?cl\<^sub>2 ! i))" by auto
      then have "(s, get_source (snd ?cl\<^sub>2 ! i)) \<in> (rstep (set R))\<^sup>*"
        using reachable_terms in_set_member by metis
      then show "(get_source (fst ?p), get_source (snd (?c ?cl\<^sub>2) ! i)) \<in> (rstep (set R))\<^sup>*" 
        using get_source_closed[OF i2, of \<box> \<rho>] \<rho>
        by (auto intro: rsteps_closed_subst simp: get_source_def)
    next
      fix i assume i:"i < length (snd (?c ?cr\<^sub>1))"
      then have i2:"i < length (snd ?cr\<^sub>1)" by auto
      then have "snd ?cr\<^sub>1 ! i \<in> set (snd ?cr\<^sub>1)" using nth_mem by auto
      with assms[unfolded Let_def check_cpeak_eldc_def] Some
      have "List.member (reachable_terms R s n) (get_source (snd ?cr\<^sub>1 ! i))" by auto
      then have "(s, get_source (snd ?cr\<^sub>1 ! i)) \<in> (rstep (set R))\<^sup>*"
        using reachable_terms in_set_member by metis
      then show "(get_source (fst ?p), get_source (snd (?c ?cr\<^sub>1) ! i)) \<in> (rstep (set R))\<^sup>*" 
       using get_source_closed[OF i2, of \<box> \<rho>] \<rho>
       by (auto intro: rsteps_closed_subst simp: get_source_def)
    next
      fix i assume i:"i < length (snd (?c ?cr\<^sub>2))"
      then have i2:"i < length (snd ?cr\<^sub>2)" by auto
      then have "snd ?cr\<^sub>2 ! i \<in> set (snd ?cr\<^sub>2)" using nth_mem by auto
      with assms[unfolded Let_def check_cpeak_eldc_def] Some
      have "List.member (reachable_terms R s n) (get_source (snd ?cr\<^sub>2 ! i))" by auto
      then have "(s, get_source (snd ?cr\<^sub>2 ! i)) \<in> (rstep (set R))\<^sup>*"
        using reachable_terms in_set_member by metis
      then show "(get_source (fst ?p), get_source (snd (?c ?cr\<^sub>2) ! i)) \<in> (rstep (set R))\<^sup>*" 
        using get_source_closed[OF i2, of \<box> \<rho>] \<rho>
        by (auto intro: rsteps_closed_subst simp: get_source_def)
    next
      fix i assume i:"i < length (snd (?s ?sl))"
      then have i2:"i < length (snd ?sl)" by auto
      then have "snd ?sl ! i \<in> set (snd ?sl)" using nth_mem by auto
      with assms[unfolded Let_def check_cpeak_eldc_def] Some
      have "List.member (reachable_terms R s n) (get_source (snd ?sl ! i))" by auto
      then have "(s, get_source (snd ?sl ! i)) \<in> (rstep (set R))\<^sup>*"
        using reachable_terms in_set_member by metis
      then show "(get_source (fst ?p), get_source (snd (?s ?sl) ! i)) \<in> (rstep (set R))\<^sup>*" 
        using get_source_closed[OF i2, of \<box> \<rho>] \<rho>
        by (auto intro: rsteps_closed_subst simp: get_source_def)
    next
      fix i assume i:"i < length (snd (?s ?sr))"
      then have i2:"i < length (snd ?sr)" by auto
      then have "snd ?sr ! i \<in> set (snd ?sr)" using nth_mem by auto
      with assms[unfolded Let_def check_cpeak_eldc_def] Some
      have "List.member (reachable_terms R s n) (get_source (snd ?sr ! i))" by auto
      then have "(s, get_source (snd ?sr ! i)) \<in> (rstep (set R))\<^sup>*"
        using reachable_terms in_set_member by metis
      then show "(get_source (fst ?p), get_source (snd (?s ?sr) ! i)) \<in> (rstep (set R))\<^sup>*" 
        using get_source_closed[OF i2, of \<box> \<rho>] \<rho>
        by (auto intro: rsteps_closed_subst simp: get_source_def)
     qed
  next
    case None
    with assms show ?thesis unfolding check_cpeak_eldc_def by auto
  qed
  ultimately show ?thesis unfolding eld_fan_def eldc_def by fast
qed

lemma linear_imp_SN_rel_d_nd:
  assumes "linear_trs R"
  shows "SN_rel (rstep (R_d R)) (rstep (R_nd R))"
proof -
  from assms have "R_d R = {}" unfolding R_d_def linear_trs_def by auto
  then show ?thesis using SN_rel_empty1 by auto
qed

type_synonym ('f,'v) crit_peak_convs =  "(('f,'v) term \<times> 
  ('f,'v) term \<times> ('f, 'v) eseq \<times> ('f, 'v) rseq  \<times> ('f, 'v) eseq \<times> 
  ('f,'v) term \<times> ('f, 'v) eseq \<times> ('f, 'v) rseq  \<times> ('f, 'v) eseq) list"

context
  fixes ren :: "'v :: {infinite,showl,compare_order} renaming2" 
begin
definition check_rule_labeling_eld ::
  "('f :: {compare_order, showl}, 'v) rules \<Rightarrow>  ('f :: {compare_order, showl}, 'v) rule_lab_repr \<Rightarrow> 
     ('f :: {compare_order, showl},'v) crit_pair_joins \<Rightarrow> showsl check"
where
  "check_rule_labeling_eld R lab js = do {
  let cps = critical_peaks_impl ren R R;
  let l = rule_lab_repr_to_lab lab;
  let joins = js @ (map (\<lambda> (u, j1, v, j2). (v, j2, u ,j1)) js);
  check_allm (\<lambda>(b, (s1,r1,p1,\<sigma>1,True,t1),(s2,r2,p2,\<sigma>2,True,t2)). 
    try check (t1 = t2) (showsl_lit (STR '' pair non-trivial '')) catch
    (\<lambda>e. check_exm (\<lambda>(u, j1, v, j2). 
          (check_cpeak_eld R ((s1,r1,p1,\<sigma>1,True,t1),(s2,r2,p2,\<sigma>2,True,t2)) (u, v) j1 j2 l))
      joins (\<lambda>es. showsl_lit (STR ''\<newline>the critical peak '') \<circ> showsl_rule' showsl showsl (STR '' <- . -> '') (t1, t2) 
        \<circ> showsl_lit (STR '' could not be joined decreasingly:\<newline>'') \<circ> showsl_sep id id es)))
    cps
  }"

lemma check_rule_labeling_eld:
  assumes "isOK (check_rule_labeling_eld R lab js)"
  and "SN_rel (rstep (R_d (set R))) (rstep (R_nd (set R)))"
  and "left_linear_trs (set R)"
  shows "CR (rstep (set R))"
using assms (3,2)
proof (rule rule_labeling_is_sound_ll)
  note ok = assms(1)[unfolded check_rule_labeling_eld_def, simplified]
  let ?l = "rule_lab_repr_to_lab lab"
  let ?rl = "rule_labeling ?l"
  let ?joins = "js @ (map (\<lambda> (u, j1, v, j2). (v, j2, u ,j1)) js)"
  define r where "r = ({(n, m). n < (m :: nat)}, {(n, m). n \<le> (m :: nat)})"
  show "\<forall>(b, p) \<in> critical_peaks ren (set R). eld (set R) ?rl r p"
  proof
    fix b p
    assume p:"(b, p) \<in> critical_peaks ren (set R)"
    then have pi:"(b, p) \<in> set (critical_peaks_impl ren R R)" using critical_peaks_impl by auto
    then obtain s1 r1 p1 \<sigma>1 t1 s2 r2 p2 \<sigma>2 t2 where
      fields: "p = ((s1,r1,p1,\<sigma>1,True,t1),(s2,r2,p2,\<sigma>2,True,t2))" 
      by (cases p, auto simp: critical_peaks_impl critical_peaks_def)
    from bspec[OF ok pi]
    have "(t1 = t2) \<or> (isOK (existsM (\<lambda>(u, j1, v, j2). check_cpeak_eld R p (u, v) j1 j2 ?l) ?joins))"
      unfolding fields by simp
    then show "eld (set R) ?rl r p"
    proof
      assume "t1 = t2"
      moreover have "p \<in> local_peaks (set R)" using critical_peak_is_local_peak fields p by fast
      ultimately show ?thesis using trivial_peak_eld unfolding fields by fast
    next
      assume "isOK (existsM (\<lambda>(u, j1, v, j2). check_cpeak_eld R p (u, v) j1 j2 ?l) ?joins)"
      then obtain u j1 v j2 where "isOK(check_cpeak_eld R p (u, v) j1 j2 ?l)"
        unfolding isOK_existsM by (split prod.splits, force)
      then show ?thesis using check_cpeak_eld p unfolding r_def fields by fast
    qed
  qed
qed


definition check_rule_labeling_eldc ::
  "('f :: {compare_order, showl}, 'v) rules \<Rightarrow>  ('f :: {compare_order, showl}, 'v) rule_lab_repr \<Rightarrow> 
     ('f :: {compare_order, showl},'v) crit_peak_convs \<Rightarrow> nat option \<Rightarrow> showsl check"
where
  "check_rule_labeling_eldc R lab cs n = do {
  let cps = critical_peaks_impl ren R R;
  let l = rule_lab_repr_to_lab lab;
  let convs = cs @ (map (\<lambda> (s, u, cl\<^sub>1, sl, cl\<^sub>2, v, cr\<^sub>1, sr, cr\<^sub>2). (s, v, cr\<^sub>1, sr, cr\<^sub>2, u ,cl\<^sub>1, sl, cl\<^sub>2)) cs);
  check_allm (\<lambda>(b, (s1, r1, p1, \<sigma>1, True, t1),(s2, r2, p2, \<sigma>2, True, t2)). 
    try check (t1 = t2) (showsl_lit (STR '' pair non-trivial '')) catch
    (\<lambda>e. check_exm (\<lambda>(s, u, cl\<^sub>1, sl, cl\<^sub>2, v, cr\<^sub>1, sr, cr\<^sub>2). 
        (check_cpeak_eldc R ((s1, r1, p1, \<sigma>1, True, t1),(s2, r2, p2, \<sigma>2, True, t2)) s (u, v) cl\<^sub>1 sl cl\<^sub>2 cr\<^sub>1 sr cr\<^sub>2 l n))
      convs (\<lambda>es. showsl_nl \<circ> showsl_lit (STR ''the critical peak '') \<circ> 
        showsl t1 \<circ> showsl_lit (STR '' <- '') \<circ> showsl s1 \<circ> showsl_lit (STR '' -> '') \<circ> showsl t2 \<circ>
        showsl_lit (STR '' could not be joined decreasingly:'') \<circ> showsl_nl \<circ> showsl_sep id id es)))
    cps
  }"

lemma check_rule_labeling_eldc:
  assumes "isOK (check_rule_labeling_eldc R lab cs n)"
  and "SN_rel (rstep (R_d (set R))) (rstep (R_nd (set R)))"
  and "left_linear_trs (set R)"
  and "n = None \<longrightarrow> linear_trs (set R)"
  shows "CR (rstep (set R))"
proof -
  note ok = assms(1)[unfolded check_rule_labeling_eldc_def, simplified]
  let ?l = "rule_lab_repr_to_lab lab"
  let ?rl = "rule_labeling ?l"
  let ?convs = "cs @ (map (\<lambda> (s, u, cl\<^sub>1, sl, cl\<^sub>2, v, cr\<^sub>1, sr, cr\<^sub>2). (s, v, cr\<^sub>1, sr, cr\<^sub>2, u ,cl\<^sub>1, sl, cl\<^sub>2)) cs)"
  define r where "r = ({(n, m). n < (m :: nat)}, {(n, m). n \<le> (m :: nat)})"
  { fix b p
    assume p:"(b, p) \<in> critical_peaks ren (set R)"
    then have pi:"(b, p) \<in> set (critical_peaks_impl ren R R)" using critical_peaks_impl by auto
    then obtain s r1 p1 \<sigma>1 t1 r2 p2 \<sigma>2 t2 where fields: "p = ((s,r1,p1,\<sigma>1,True,t1),(s,r2,p2,\<sigma>2,True,t2))" 
      by (cases p, auto simp: critical_peaks_impl critical_peaks_def)
    from bspec[OF ok pi]
    consider (trivial) "(t1 = t2)" | (ok) "isOK (existsM (\<lambda>(s, u, cl\<^sub>1, sl, cl\<^sub>2, v, cr\<^sub>1, sr, cr\<^sub>2). 
      check_cpeak_eldc R p s (u, v) cl\<^sub>1 sl cl\<^sub>2 cr\<^sub>1 sr cr\<^sub>2 ?l n) ?convs)"
      unfolding fields by fastforce
    then have "eld_fan (set R) ?rl r p \<or> eldc (set R) ?rl r p \<and> linear_trs (set R)"
    proof (cases)
      case (trivial)
      moreover have "p \<in> local_peaks (set R)" using critical_peak_is_local_peak fields p by fast
      ultimately show ?thesis using trivial_peak_eld_fan unfolding fields by fast
    next
      case (ok)
      then obtain s' u cl\<^sub>1 sl cl\<^sub>2 v cr\<^sub>1 sr cr\<^sub>2 where "isOK(check_cpeak_eldc R p s' (u, v) cl\<^sub>1 sl cl\<^sub>2 cr\<^sub>1 sr cr\<^sub>2 ?l n)"
        unfolding isOK_existsM by (split prod.splits, force)
      from check_cpeak_eldc[OF p[unfolded fields] this[unfolded fields]] assms(4) show ?thesis 
        unfolding r_def fields by auto
    qed
  }
  then consider (fan) "(\<forall>(b, p) \<in> critical_peaks ren (set R). eld_fan (set R) ?rl r p)"
    | (linear) "(\<forall>(b, p) \<in> critical_peaks ren (set R). eldc (set R) ?rl r p) \<and> linear_trs (set R)"
    unfolding eld_fan_def eldc_def by blast
  then show ?thesis
  proof (cases)
    case fan
    then show ?thesis using rule_labeling_conv_is_sound_ll[OF assms(3,2)] unfolding r_def by auto
  next
    case linear
    then show ?thesis using rule_labeling_is_sound unfolding r_def by auto
  qed
qed

end
end
